/*
 
 *****************************************************************************
 * Author:                                                                   *
 * ------                                                                    *
 *  Anton Kokalj                                  Email: Tone.Kokalj@ijs.si  *
 *  Department of Physical and Organic Chemistry  Phone: x 386 1 477 3523    *
 *  Jozef Stefan Institute                          Fax: x 386 1 477 3811    *
 *  Jamova 39, SI-1000 Ljubljana                                             *
 *  SLOVENIA                                                                 *
 *                                                                           *
 * Source: $XCRYSDEN_TOPDIR/C/xcAtomAdd.c
 * ------                                                                    *
 * Copyright (c) 1996-2003 by Anton Kokalj                                   *
 *****************************************************************************

*/

#include "togl.h" 
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <tk.h>
#include <GL/glu.h>
#include "struct.h"
#include "displayfunc.h"
#include "vector.h"
#include "xcfunc.h"
#include "memory.h"

#define ADDLINE_PAT 0xAAAA  /* addline stipple pattern */
#define ADDLINE_WIDTH 2.0   /* width of "addline" */


GLuint BasicVectorsList, AtomAddList;
GLdouble arrcoor[2][3][3], veccoor[2][3][3];
GLdouble points[8][3];
GLdouble addatompos[3];
extern GLuint tempDisable3Dlist;
extern GLuint tempEnable3Dlist;


/* --- FUNCTIONS PROTOTYPES --- */
int XC_AtomAddCmd(ClientData clientData, Tcl_Interp *interp,
		  int argc, char *argv[]);
static void SetBasicVectors( double vec[][4], int n );
static char *xcAddAtomUpdate(double af, double bf, double cf);
void xcDisplayAddAtomBox(void);
/*static void xcMakeAtomAddList(void);*/


/* --- xcDisplayFunc.c --- */
extern void xcSolidCylinder (GLdouble radius, GLdouble height);
extern void xcSolidCone (GLdouble baseradius, GLdouble topradius, 
			 GLdouble height);
/* extern void (*xcDisplay)(void); */

/* --- xcSelect.c --- */
extern void GetSelCylinderPar(double x21, double y21, double z21, 
			      double *xrvb, double *yrvb, double *zrvb, 
			      double *fibond, double *bondl);



/* XC_AtomAddCmd --> inplementation of 'xc_atomadd' custom Tcl command,
                     which is used for Cell-Adding type of ATOMINSE
		     crystal command
 * ----------------------------------------------------------------------------
 * Usage: xc_atomadd <toglName> begin|update|clean ?AF? ?BF? ?CF? 
 *		
 *              xc_atomadd begin -- before we begin atomadd procedure 
 *                                  we must prepare and do some initialisations
 *
 *              xc_atomadd update-- updating
 *
 *              xc_atomadd clean -- clean-up (distroy BasicVectorsList)
 */	
int 
XC_AtomAddCmd(ClientData clientData, Tcl_Interp *interp,
	      int argc, char *argv[])
{
  struct Togl *togl;
  char *csqn = (char*) Tcl_Alloc( sizeof(char) * 40 );
  GLdouble af,bf,cf; /* factors for vectors AF BF CF */

  if (argc < 3) {
    interp->result = "Usage: xc_atomadd <toglName> begin|update|clean ?AF? ?BF? ?CF?";
    return TCL_ERROR;
  }

  /* find togl associated with toglName */
  if ( ( togl = Togl_FindTogl(argv[1]) ) == NULL ) {
    sprintf(interp->result, 
	    "couldn't find %s togl widget", argv[1]);
    return TCL_ERROR;
  }

  /* if structure is not opened, just return silently */
  if ( !VPf.stropened ) return TCL_OK;

  /* so far just XC_3D is implemented; if not XC_3D return silently */
  if ( dimType != XC_3D ) return TCL_OK;

  /***************************************************************************/
  /* first "XC_ATOMADD <toglName> BEGIN" */
  if ( strcmp(argv[2],"begin") == 0 ) {
    int n;
    char *csqn = (char*) Tcl_Alloc( sizeof(char) * 10 );
    /* SO FAR ONLY 3D XC_ATOMADD IMPLEMENTATION */
    /* flag to know when we are in "atomadd mode" */
    VPf.atomadd = GL_TRUE;

    /* 3D ONLY: make basic vectors appear on the screen */    
    /* if xcr.lprimvec == GL_TRUE --> deal with primitiv vectors */
    /* if xcr.lconvvec == GL_TRUE --> deal with primitiv vectors */
    /* if xcr.dim  == 3 --> crystal
     *             == 2 --> SLAB; put C vector to (0.0,0.0,1.0)
     *		   == 1 --> polymer; put B,C vectors to 
     *                               (0.0,(1.0|0.0),(0.0|1.0))
     */
    n = xcr.dim;
    if ( xcr.celltype == XCR_PRIMCELL ) SetBasicVectors( vec.prim, n ); 
    if ( xcr.celltype == XCR_CONVCELL ) SetBasicVectors( vec.conv, n ); 
    /* make a box for marking current atomadd position */
    /* xcMakeAtomAddList(); */
    /* default values for A B C fractions will be 0.3 0.3 0.3, that is what we
       will return */
    af = 0.3;
    bf = 0.3;
    cf = 0.3;
    /* now update and return current position */
    csqn = xcAddAtomUpdate(af, cf, bf);
    Tcl_SetResult(interp,csqn,TCL_DYNAMIC);
    /* now update display */
    xcDisplayAddAtomBox();
    glFlush();
    Togl_SwapBuffers(togl);
  }

  /***************************************************************************/
  /* now implement a "XC_ATOMADD <toglName> UPDATE %AF %BF %CF"; 
     %AF,%BF,%CF are fractions of AF BF CF crystal vectors */
  /* at the end we will calculate current position */
  else if ( strcmp(argv[2],"update") == 0 ) {

    if ( !VPf.atomadd ) {
      interp->result = "calling \"xc_atomadd <toglName> update\" command, before initialising atom-add process with \"xc_atomadd <toglName> begin\"";
      return TCL_ERROR;
    }    

    /* argc must be 5 */
    if ( argc != 6 ) {
      interp->result = "wrong usage of \"xc_atomadd <toglName> update\" command, must be \"xc_addatom <toglName> update <AF> <BF> <CF>\"";
      return TCL_ERROR;
    }
    /* argv[3],argv[4],argv[5] must be integers */
    if ( Tcl_GetDouble(interp,argv[3],&af) == TCL_ERROR ) {
      sprintf(interp->result,"got invalid value for <AF>, while executing %s %s %s %s %s %s\n",argv[0],argv[1],argv[2],argv[3],argv[4],argv[5]);
      return TCL_ERROR;
    }
    if ( Tcl_GetDouble(interp,argv[4],&bf) == TCL_ERROR ) {
      sprintf(interp->result,"got invalid value for <BF>, while executing %s %s %s %s %s %s\n",argv[0],argv[1],argv[2],argv[3],argv[4],argv[5]);
      return TCL_ERROR;
    }
    if ( Tcl_GetDouble(interp,argv[5],&cf) == TCL_ERROR ) {
      sprintf(interp->result,"got invalid value for <CF>, while executing %s %s %s %s %s %s\n",argv[0],argv[1],argv[2],argv[3],argv[4],argv[5]);
      return TCL_ERROR;
    }

    /* now update and return current position */
    csqn = xcAddAtomUpdate(af, bf, cf);
    Tcl_SetResult(interp,csqn,TCL_DYNAMIC);
    /* update the display */
    Togl_PostRedisplay(togl);
  }

  /***************************************************************************/
  /* XC_ATOMADD <toglName> CLEAN */
  else if ( strcmp(argv[2],"clean") == 0 ) {
    if ( !VPf.atomadd ) {
      interp->result = "calling \"xc_atomadd <toglName> clean\" command, before initialising atom-add process with \"xc_atomadd <toglName> begin\"";
      return TCL_ERROR;      
    }    
    /* destroy BasicVectorsList */
    glDeleteLists( BasicVectorsList, 1 );
    VPf.atomadd = GL_FALSE;
    /* now update display */
    Togl_PostRedisplay(togl);
  }
  /***************************************************************************/
  /* unkown option */
  else {
    sprintf(interp->result,"unkown option \"%s\" for \"xc_atomadd\" comand, must be one of begin, update, clean",argv[2]);
    return TCL_ERROR;
  }

  return TCL_OK;
}


/*****************************************************************************/
static void 
SetBasicVectors( double vec[][4], int n )
{
  int i,j;
  double mxx, myy, mzz;
  GLdouble vecthick, arrthick;
  GLdouble arrx[3], arry[3], arrz[3], arrfi[3], arrl[3];
  GLdouble vecx[3], vecy[3], vecz[3], vecfi[3], vecl[3];

  /* for polymer & SLAB modify (B) and C vector to 
     (0.0,(1.0|0.0),(0.0|1.0)) */
  for (i = 0; i<=2; i++) {
    if ( i >= n ) {
      vec[i][i] = 1.0; /* modified is just part that should be 1.0 */
    }
    /* arrow will be 0.25 of vector length */
    for (j = 0; j<=2; j++) {
      veccoor[0][i][j] = 0.0;
      veccoor[1][i][j] = 0.75 * vec[i][j];
      arrcoor[0][i][j] = 0.75 * vec[i][j];
      arrcoor[1][i][j] = vec[i][j];
    }
    /* convert cartesian coordinates to gluCylinder coor. */
    GetSelCylinderPar(veccoor[1][i][0] - veccoor[0][i][0], 
		      veccoor[1][i][1] - veccoor[0][i][1],
		      veccoor[1][i][2] - veccoor[0][i][2],
		      vecx + i, vecy + i, vecz + i,
		      vecfi + i, vecl + i);
    GetSelCylinderPar(arrcoor[1][i][0] - arrcoor[0][i][0], 
		      arrcoor[1][i][1] - arrcoor[0][i][1],
		      arrcoor[1][i][2] - arrcoor[0][i][2],
		      arrx + i, arry + i, arrz + i,
		      arrfi + i, arrl + i);
  }
  
  /* vector will be 1.3 times thicker than bond & arrow will be 1.5 times 
     thicker than vector */
  vecthick = VECTOR_THICKF * rrod;
  arrthick = VECTOR_ARRTHICKF * vecthick;
  
  mxx = mx;
  myy = my;
  mzz = mz;
  /**************************************************************************/
  /* maybe for MOLECULE it's better to draw vectors from MASS-CENTER,       */
  /* because of the presence of point-groups                                */
  /**************************************************************************/
  if ( xcr.dim == 0 ) {
    mxx = 0.0;
    myy = 0.0;
    mzz = 0.0;
  }
  /* what if list already exists ?????????????????? */
  glNewList(BasicVectorsList, GL_COMPILE_AND_EXECUTE);
     glTranslated( -mxx, -myy, -mzz ); /* (mx,my,mz) is origin shift vector */
     glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, 
		  atcol[0] );
     for (i=0; i<=2; i++) {      
       /* first rod of vector */    
       glRotated( vecfi[i], vecx[i], vecy[i], vecz[i] );
       xcSolidCylinder(vecthick, vecl[i]);
       glRotated( -vecfi[i], vecx[i], vecy[i], vecz[i] );
       /* now arrow of vector */
       glTranslated( arrcoor[0][i][0], arrcoor[0][i][1], arrcoor[0][i][2] );
       glRotated( arrfi[i], arrx[i], arry[i], arrz[i] );
       xcSolidCone(arrthick, 0.0, arrl[i]);
       glRotated( -arrfi[i], arrx[i], arry[i], arrz[i] );
       glTranslated( -arrcoor[0][i][0], -arrcoor[0][i][1], -arrcoor[0][i][2] );
     }
     glTranslated( mxx, myy, mzz );
  glEndList();
}


/*****************************************************************************/
/* CALCULATE NEW CURRENT ADDATOM POSITION & RETURN IT */
static char 
*xcAddAtomUpdate(double af, double bf, double cf)
{  
  int i;
  char *addatomdata = (char *) xcMalloc(sizeof(char) * 160);

  /* points for cube-box edges */
  for(i=0; i<=2; i++) {
    points[0][i] = 0.0;
    points[1][i] = af*arrcoor[1][0][i];
    points[2][i] = af*arrcoor[1][0][i] + bf*arrcoor[1][1][i];
    points[3][i] = bf*arrcoor[1][1][i];
    points[4][i] = cf*arrcoor[1][2][i];
    points[5][i] = af*arrcoor[1][0][i] + cf*arrcoor[1][2][i];
    points[6][i] = points[5][i] + bf*arrcoor[1][1][i];
    points[7][i] = bf*arrcoor[1][1][i] + cf*arrcoor[1][2][i];
    addatompos[i] = af*arrcoor[1][0][i] + bf*arrcoor[1][1][i]
                    + cf*arrcoor[1][2][i];
  }    

  /* return the addatom current position */
  return addatomdata;
}


void
xcDisplayAddAtomBox(void)
{
  double mxx, myy, mzz;

  mxx = mx;
  myy = my;
  mzz = mz;
  /**************************************************************************/
  /* maybe for MOLECULE it's better to draw vectors from MASS-CENTER,       */
  /* because of the presence of point-groups                                */
  /**************************************************************************/
  if ( xcr.dim == 0 ) {
    mxx = 0.0;
    myy = 0.0;
    mzz = 0.0;
  }

  glLineStipple(2, ADDLINE_PAT);
  glEnable(GL_LINE_STIPPLE);
  glCallList(tempDisable3Dlist);
  glLineWidth( ADDLINE_WIDTH );
  /* make a addatombox */
  /* box is made from this poits:
   *                              01,12,23,30
   *                              04,15,26,37
   *                              45,56,67,74
   */
  glTranslated( -mxx, -myy, -mzz ); /* (mx,my,mz) is origin shift vector */
  glBegin(GL_LINE_STRIP);
    glVertex3dv( points[0]);
    glVertex3dv( points[1]);
    glVertex3dv( points[2]);
    glVertex3dv( points[3]);
    glVertex3dv( points[0]);
  glEnd();
  glBegin(GL_LINES);
    glVertex3dv( points[0]);
    glVertex3dv( points[4]);
    glVertex3dv( points[1]);
    glVertex3dv( points[5]);
    glVertex3dv( points[2]);
    glVertex3dv( points[6]);
    glVertex3dv( points[3]);
    glVertex3dv( points[7]);
  glEnd();
  glBegin(GL_LINE_STRIP);
    glVertex3dv( points[4]);
    glVertex3dv( points[5]);
    glVertex3dv( points[6]);
    glVertex3dv( points[7]);
    glVertex3dv( points[4]);
  glEnd();
  /* make a point on atomadd-position */
  glPointSize(6);
  glBegin(GL_POINTS);
    glVertex3dv( addatompos );
  glEnd();
  glTranslated( mxx, myy, mzz );

  glCallList(tempEnable3Dlist);
  glDisable(GL_LINE_STIPPLE);
} 


/*
static void 
xcMakeAtomAddList(void) 
{
  float color[3] = {1.0, 1.0, 1.0};
  GLUquadricObj *quadObj;
  quadObj = gluNewQuadric ();
  glNewList(AtomAddList, GL_COMPILE);
    glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, 
		  color);
    gluQuadricDrawStyle (quadObj, GLU_FILL);
    gluQuadricNormals (quadObj, GLU_FLAT);
    gluSphere (quadObj, 0.1, 4, 4);
  glEndList();
}
*/
